"""
Code providing piecewise quadratic bounds on the logistic log partition
function and methods to compute the expectations of this bound with
respect to a univariate Gaussian.

This is based on Matlab code written by Emtiyaz Khan
(www.cs.ubc.ca/~emtiyaz), much of it being a direct translation of his
code into python.
"""

from __future__ import division

__author__ = 'Matthew W. Hoffman'
__date__ = '2012/4/4'

import numpy as np
from scipy.stats.distributions import norm as normal

__all__ = ['get_expectation', 'get_bound']

def get_expectation(m, v, (p, a)):
    """
    Get the expectations and gradients of the piecewise quadratic bound
    given by the tuple (p,a) under some number of Gaussians with means m
    and variances v.

    This will return (fr, gm, gv) containing the expectations,
    derivatives with respect to m, and derivatives with respect to v.
    Finally, if a 1d array of means/variances is passed, these
    quantities will be arrays of the same size.
    """
    m = np.asarray(m)
    v = np.asarray(v)
    sqrt_v = np.sqrt(v)

    c, b, a = map(np.reshape, a.T, ((-1,1),)*3)
    l = np.r_[-np.inf, p].reshape(-1,1)
    h = np.r_[ p, np.inf].reshape(-1,1)

    # these quantities, and hence all the derived quantities, will be
    # arrays of size (r,d).
    zl = (l - m) / sqrt_v
    zh = (h - m) / sqrt_v

    pl = normal.pdf(zl) / sqrt_v
    ph = normal.pdf(zh) / sqrt_v
    cl = normal.cdf(zl)
    ch = normal.cdf(zh)

    # the only time we will use the values of the bounds below is when
    # they are being multiplied by pl/ph respectively. So rather than
    # using inf*0.0 and getting NaN, we'll just set these to zero.
    l[0] = h[-1] = 0.0

    # truncated moments.
    ex0 = ch - cl
    ex1 = v*(pl-ph) + m*(ch-cl)
    ex2 = ((l+m)*pl - (h+m)*ph) * v + (ch - cl) * (v+m**2)
    fr = a*ex2 + b*ex1 + c*ex0

    gm = a*((l**2 + 2*v)*pl - (h**2 + 2*v)*ph + 2*m*(ch-cl)) + \
         b*(l*pl - h*ph + ch - cl) + c*(pl - ph)

    gv = (0.5 * a / v) * ((2*l*v + l**3 - l**2*m)*pl - (2*h*v + h**3 - h**2*m)*ph) + a*(ch - cl) + \
         (0.5 * b / v) * ((l**2 + v - l*m)*pl - (h**2 + v - h*m)*ph) + \
         (0.5 * c / v) * ((l - m)*pl - (h - m)*ph)

    return fr.sum(axis=0), gm.sum(axis=0), gv.sum(axis=0)

def get_bound(r):
    """
    Return components describing a piecewise quadratic bound on the
    logistic log-partition function with r pieces.

    This will return a tuple (p,a) where p is an (r-1)-array containing
    the boundaries of the pieces (there are two implicit boundaries at
    -inf/inf) and a is an (r,3)-array describing the multiplicative
    constants for each piece.
    """
    if r == 3:
        p = np.array([-2.950292e+00, 2.950292e+00])
        a = np.array([
            [5.100147e-02, 0.000000e+00, 0.000000e+00, ],
            [7.441486e-01, 5.000000e-01, 9.570071e-02, ],
            [5.100147e-02, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 4:
        p = np.array([-3.396125e+00, 0.000000e+00, 3.396125e+00])
        a = np.array([
            [3.295385e-02, 0.000000e+00, 0.000000e+00, ],
            [6.931472e-01, 4.159071e-01, 6.808183e-02, ],
            [6.931472e-01, 5.840929e-01, 6.808183e-02, ],
            [3.295385e-02, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 5:
        p = np.array(
            [-4.136996e+00, -1.134605e+00, 1.134605e+00,
              4.136996e+00])
        a = np.array([
            [1.584457e-02, 0.000000e+00, 0.000000e+00, ],
            [5.388657e-01, 2.694962e-01, 3.550907e-02, ],
            [6.970314e-01, 5.016896e-01, 1.180782e-01, ],
            [5.388657e-01, 7.305038e-01, 3.550907e-02, ],
            [1.584457e-02, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 6:
        p = np.array(
            [-4.832591e+00, -1.967034e+00, 0.000000e+00,
              1.967034e+00,  4.832591e+00])
        a = np.array([
            [7.934298e-03, 0.000000e+00, 0.000000e+00, ],
            [3.841260e-01, 1.647533e-01, 1.831495e-02, ],
            [6.967717e-01, 4.794020e-01, 9.945062e-02, ],
            [6.967717e-01, 5.205980e-01, 9.945062e-02, ],
            [3.841260e-01, 8.352467e-01, 1.831495e-02, ],
            [7.934298e-03, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 7:
        p = np.array(
            [-5.305268e+00, -2.479972e+00, -7.684416e-01,
              7.684416e-01,  2.479972e+00,  5.305268e+00])
        a = np.array([
            [4.953082e-03, 0.000000e+00, 0.000000e+00, ],
            [2.956403e-01, 1.156815e-01, 1.165311e-02, ],
            [6.488156e-01, 4.015220e-01, 7.029358e-02, ],
            [6.950378e-01, 4.987377e-01, 1.205014e-01, ],
            [6.488156e-01, 5.984780e-01, 7.029358e-02, ],
            [2.956403e-01, 8.843185e-01, 1.165311e-02, ],
            [4.953082e-03, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 8:
        p = np.array(
            [-5.723949e+00, -2.963782e+00, -1.341209e+00,
              0.000000e+00,  1.341209e+00,  2.963782e+00,
              5.723949e+00])
        a = np.array([
            [3.261461e-03, 0.000000e+00, 0.000000e+00, ],
            [2.216441e-01, 7.936676e-02, 7.276484e-03, ],
            [5.707796e-01, 3.167044e-01, 4.798202e-02, ],
            [6.931472e-01, 4.896596e-01, 1.100816e-01, ],
            [6.931472e-01, 5.103404e-01, 1.100816e-01, ],
            [5.707796e-01, 6.832956e-01, 4.798202e-02, ],
            [2.216441e-01, 9.206332e-01, 7.276484e-03, ],
            [3.261461e-03, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 9:
        p = np.array(
            [-6.106671e+00, -3.355650e+00, -1.805672e+00,
             -5.979038e-01,  5.979038e-01,  1.805672e+00,
              3.355650e+00,  6.106671e+00])
        a = np.array([
            [2.225477e-03, 0.000000e+00, 0.000000e+00, ],
            [1.743334e-01, 5.861411e-02, 5.030641e-03, ],
            [4.930632e-01, 2.501563e-01, 3.400341e-02, ],
            [6.769707e-01, 4.523365e-01, 9.020235e-02, ],
            [6.935075e-01, 4.995219e-01, 1.232522e-01, ],
            [6.769707e-01, 5.476635e-01, 9.020235e-02, ],
            [4.930632e-01, 7.498437e-01, 3.400341e-02, ],
            [1.743334e-01, 9.413859e-01, 5.030641e-03, ],
            [2.225476e-03, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 10:
        p = np.array(
            [-6.447793e+00, -3.713484e+00, -2.194193e+00,
             -1.030626e+00,  0.000000e+00,  1.030626e+00,
              2.194193e+00,  3.713484e+00,  6.447793e+00])
        a = np.array([
            [1.582760e-03, 0.000000e+00, 0.000000e+00, ],
            [1.379276e-01, 4.381455e-02, 3.544366e-03, ],
            [4.213367e-01, 1.977651e-01, 2.456452e-02, ],
            [6.392300e-01, 3.962588e-01, 7.009680e-02, ],
            [6.932852e-01, 4.956213e-01, 1.161149e-01, ],
            [6.932852e-01, 5.043787e-01, 1.161149e-01, ],
            [6.392300e-01, 6.037412e-01, 7.009680e-02, ],
            [4.213367e-01, 8.022349e-01, 2.456452e-02, ],
            [1.379276e-01, 9.561854e-01, 3.544366e-03, ],
            [1.582760e-03, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 15:
        p = np.array(
            [-7.702428e+00, -5.033256e+00, -3.592218e+00,
             -2.561523e+00, -1.732214e+00, -1.001394e+00,
             -3.310443e-01,  3.310443e-01,  1.001394e+00,
              1.732214e+00,  2.561523e+00,  3.592218e+00,
              5.033256e+00,  7.702428e+00])
        a = np.array([
            [4.516270e-04, 0.000000e+00, 0.000000e+00, ],
            [5.507571e-02, 1.458683e-02, 9.805122e-04, ],
            [2.008333e-01, 7.282215e-02, 6.813733e-03, ],
            [3.884465e-01, 1.777045e-01, 2.150595e-02, ],
            [5.567715e-01, 3.093106e-01, 4.729739e-02, ],
            [6.583594e-01, 4.261014e-01, 8.099695e-02, ],
            [6.909407e-01, 4.890991e-01, 1.116390e-01, ],
            [6.932172e-01, 5.000661e-01, 1.239939e-01, ],
            [6.909407e-01, 5.109009e-01, 1.116390e-01, ],
            [6.583594e-01, 5.738986e-01, 8.099695e-02, ],
            [5.567715e-01, 6.906894e-01, 4.729739e-02, ],
            [3.884465e-01, 8.222955e-01, 2.150595e-02, ],
            [2.008333e-01, 9.271778e-01, 6.813733e-03, ],
            [5.507571e-02, 9.854132e-01, 9.805122e-04, ],
            [4.516270e-04, 1.000000e+00, 0.000000e+00, ],
        ])

    elif r == 20:
        p = np.array(
            [-8.575195e+00, -5.933689e+00, -4.525934e+00,
             -3.528108e+00, -2.751549e+00, -2.097899e+00, -1.519691e+00,
             -9.895334e-01, -4.914731e-01,  0.000000e+00,  4.914731e-01,
              9.895334e-01,  1.519691e+00,  2.097899e+00,  2.751549e+00,
              3.528108e+00,  4.525934e+00,  5.933689e+00,  8.575195e+00])
        a = np.array([
            [1.887122e-04, 0.000000e+00, 0.000000e+00, ],
            [2.809031e-02, 6.648615e-03, 3.977911e-04, ],
            [1.102118e-01, 3.443268e-02, 2.753101e-03, ],
            [2.327364e-01, 8.870197e-02, 8.770187e-03, ],
            [3.725247e-01, 1.680242e-01, 2.003476e-02, ],
            [5.045679e-01, 2.640329e-01, 3.751160e-02, ],
            [6.062803e-01, 3.607558e-01, 6.054303e-02, ],
            [6.661254e-01, 4.390945e-01, 8.625678e-02, ],
            [6.893343e-01, 4.850918e-01, 1.092135e-01, ],
            [6.931472e-01, 4.994192e-01, 1.230261e-01, ],
            [6.931472e-01, 5.005808e-01, 1.230261e-01, ],
            [6.893343e-01, 5.149082e-01, 1.092135e-01, ],
            [6.661254e-01, 5.609055e-01, 8.625678e-02, ],
            [6.062803e-01, 6.392442e-01, 6.054303e-02, ],
            [5.045679e-01, 7.359671e-01, 3.751160e-02, ],
            [3.725247e-01, 8.319758e-01, 2.003476e-02, ],
            [2.327364e-01, 9.112980e-01, 8.770187e-03, ],
            [1.102118e-01, 9.655673e-01, 2.753101e-03, ],
            [2.809031e-02, 9.933514e-01, 3.977911e-04, ],
            [1.887120e-04, 1.000000e+00, 0.000000e+00, ],
        ])

    else:
        raise ValueError('no values for r=%d piece(s)' % r)

    return p, a

